declare type NotEmpty = Exclude<any, null | false | undefined | 0 | ''>;
export declare type AsyncResultPending = {
    state: 'pending';
    isEmpty: false;
    isPending: true;
    isError: false;
    isDone: false;
    [Symbol.toStringTag]: 'AsyncResultPending';
};
export declare type AsyncResultError = {
    state: 'error';
    isEmpty: false;
    isPending: false;
    isError: true;
    isDone: false;
    error: NotEmpty;
    [Symbol.toStringTag]: 'AsyncResultError';
};
export declare type AsyncResultDone<R = unknown> = {
    state: 'done';
    isEmpty: false;
    isPending: false;
    isError: false;
    isDone: true;
    result: R;
    [Symbol.toStringTag]: 'AsyncResultDone';
};
export declare type AsyncResultEmpty = {
    state: 'empty';
    isEmpty: true;
    isPending: false;
    isError: false;
    isDone: false;
    [Symbol.toStringTag]: 'AsyncResultEmpty';
};
export declare type AsyncResult<R> = (AsyncResultPending | AsyncResultError | AsyncResultDone<R> | AsyncResultEmpty);
export declare type AsyncResultNotDone = (AsyncResultPending | AsyncResultError | AsyncResultEmpty);
export declare const AsyncResult: {
    /**
     * Returns `true` if `obj` is an AsyncResult, otherwise `false`.
     */
    isAsyncResult<R = unknown>(obj: any): obj is AsyncResult<R>;
    /**
     * Wraps a promise, ensuring that - if it completes - the result will be
     * an `AsyncResult`.
     *
     * If it resolves, the result will be an `AsyncResultDone`, if it errors
     * the result will be an `AsyncResultError`, and if it resolves to an
     * `AsyncResult`, the result will be that `AsyncResult`.
     *
     * For example:
     *
     * ``ts
     * > AsyncResult.wrap(new Promise(res => resolve("done!"))).then(console.log)
     * { state: 'done', isDone: true, ..., result: 'done!' }
     * > AsyncResult.wrap(new Promise((_, rej) => rej("error!"))).then(console.log)
     * { state: 'error', isError: true, ..., error: 'error!' }
     * > AsyncResult.wrap(new Promise(res => res(AsyncResult.empty()))).then(console.log)
     * { state: 'empty', isEmpty: true, ... }
     */
    wrap<R_1>(promise: Promise<AsyncResultPending | AsyncResultError | AsyncResultEmpty | R_1 | AsyncResultDone<R_1>>): Promise<AsyncResult<R_1>>;
    /**
     * Returns a "pending" AsyncResult.
     */
    pending(): AsyncResultPending;
    /**
     * Returns an "error" AsyncResult.
     */
    error(error: any): AsyncResultError;
    /**
     * Returns a "done" AsyncResult.
     */
    done<R_2>(res: R_2): AsyncResultDone<R_2>;
    /**
     * Returns an "empty" AsyncResult.
     */
    empty(): AsyncResultEmpty;
};
declare type UseAsyncResultBase<R> = {
    /**
     * Clears the active promise, returning the state to `AsyncResultEmpty`.
     */
    clear(): void;
    /**
     * Sets or replaces the active promise to `promise`.
     */
    bind(promise: Promise<R>): void;
};
export declare type UseAsyncResultPending<R> = UseAsyncResultBase<R> & AsyncResultPending;
export declare type UseAsyncResultError<R> = UseAsyncResultBase<R> & AsyncResultError;
export declare type UseAsyncResultDone<R> = UseAsyncResultBase<R> & AsyncResultDone<R>;
export declare type UseAsyncResultEmpty<R> = UseAsyncResultBase<R> & AsyncResultEmpty;
export declare type UseAsyncResult<R> = (UseAsyncResultPending<R> | UseAsyncResultError<R> | UseAsyncResultDone<R> | UseAsyncResultEmpty<R>);
/**
 * Expose a Promise's states (loading/error/done/empty) to a React component.
 *
 * For example:
 *
 * ```tsx
 * const userReq = useAsyncResult<User>(async () => {
 *   const resp = fetch(`/users/{userId}`)
 *   return resp.json()
 * }, [userId])
 *
 * if (userReq.isError)
 *   return <div>Error: {userReq.error}</div>
 *
 * if (userReq.isPending)
 *   return <div>Loading...</div>
 *
 * return <div>Username: {userReq.result.username}</div>
 * ```
 *
 * In addition to providing a Promise callback, the `.bind(...)` method can
 * be used to add or replace the active Promise:
 *
 * ```tsx
 * const usrSaveReq = useAsyncResult()
 * const saveUser = (newUser: User) => {
 *   userSaveReq.bind(fetch({
 *     method: 'POST',
 *     url: `/users/{userId}`,
 *     data: newUser,
 *   }))
 * }
 *
 * // ... snip ...
 *
 * return <div>
 *   ...
 *   <button
 *     disabled={saveUserReq.isPending}
 *     onClick={() => saveUser({ ... })}>
 *       Save User
 *   </button>
 *   {userSaveReq.isError && <div>Save error: {'' + userSaveReq.error}</div>}
 * </div>
 * ```
 */
export declare function useAsyncResult<R = unknown>(getPromise?: () => Promise<R | AsyncResult<R>> | AsyncResult<R> | null, useEffectOnChangeList?: any[]): UseAsyncResult<R>;
export {};
