import { useState, useEffect } from 'react';

function _extends() {
  _extends = Object.assign ? Object.assign.bind() : function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };
  return _extends.apply(this, arguments);
}

var AsyncResult = {
  /**
   * Returns `true` if `obj` is an AsyncResult, otherwise `false`.
   */
  isAsyncResult: function isAsyncResult(obj) {
    return !!obj && typeof obj.state == 'string' && typeof obj.isEmpty == 'boolean' && typeof obj.isPending == 'boolean' && typeof obj.isError == 'boolean' && typeof obj.isDone == 'boolean';
  },

  /**
   * Wraps a promise, ensuring that - if it completes - the result will be
   * an `AsyncResult`.
   *
   * If it resolves, the result will be an `AsyncResultDone`, if it errors
   * the result will be an `AsyncResultError`, and if it resolves to an
   * `AsyncResult`, the result will be that `AsyncResult`.
   *
   * For example:
   *
   * ``ts
   * > AsyncResult.wrap(new Promise(res => resolve("done!"))).then(console.log)
   * { state: 'done', isDone: true, ..., result: 'done!' }
   * > AsyncResult.wrap(new Promise((_, rej) => rej("error!"))).then(console.log)
   * { state: 'error', isError: true, ..., error: 'error!' }
   * > AsyncResult.wrap(new Promise(res => res(AsyncResult.empty()))).then(console.log)
   * { state: 'empty', isEmpty: true, ... }
   */
  wrap: function wrap(promise) {
    return promise.then(function (res) {
      return AsyncResult.isAsyncResult(res) ? res : AsyncResult.done(res);
    }, function (err) {
      return AsyncResult.isAsyncResult(err) ? err : AsyncResult.error(err);
    });
  },

  /**
   * Returns a "pending" AsyncResult.
   */
  pending: function pending() {
    var _ref;

    return _ref = {
      state: 'pending',
      isEmpty: false,
      isPending: true,
      isError: false,
      isDone: false
    }, _ref[Symbol.toStringTag] = 'AsyncResultPending', _ref;
  },

  /**
   * Returns an "error" AsyncResult.
   */
  error: function error(_error) {
    var _ref2;

    return _ref2 = {
      state: 'error',
      isEmpty: false,
      isPending: false,
      isError: true,
      isDone: false,
      error: _error
    }, _ref2[Symbol.toStringTag] = 'AsyncResultError', _ref2;
  },

  /**
   * Returns a "done" AsyncResult.
   */
  done: function done(res) {
    var _ref3;

    return _ref3 = {
      state: 'done',
      isEmpty: false,
      isPending: false,
      isError: false,
      isDone: true,
      result: res
    }, _ref3[Symbol.toStringTag] = 'AsyncResultDone', _ref3;
  },

  /**
   * Returns an "empty" AsyncResult.
   */
  empty: function empty() {
    var _ref4;

    return _ref4 = {
      state: 'empty',
      isEmpty: true,
      isPending: false,
      isError: false,
      isDone: false
    }, _ref4[Symbol.toStringTag] = 'AsyncResultEmpty', _ref4;
  }
};
/**
 * Expose a Promise's states (loading/error/done/empty) to a React component.
 *
 * For example:
 *
 * ```tsx
 * const userReq = useAsyncResult<User>(async () => {
 *   const resp = fetch(`/users/{userId}`)
 *   return resp.json()
 * }, [userId])
 *
 * if (userReq.isError)
 *   return <div>Error: {userReq.error}</div>
 *
 * if (userReq.isPending)
 *   return <div>Loading...</div>
 *
 * return <div>Username: {userReq.result.username}</div>
 * ```
 *
 * In addition to providing a Promise callback, the `.bind(...)` method can
 * be used to add or replace the active Promise:
 *
 * ```tsx
 * const usrSaveReq = useAsyncResult()
 * const saveUser = (newUser: User) => {
 *   userSaveReq.bind(fetch({
 *     method: 'POST',
 *     url: `/users/{userId}`,
 *     data: newUser,
 *   }))
 * }
 *
 * // ... snip ...
 *
 * return <div>
 *   ...
 *   <button
 *     disabled={saveUserReq.isPending}
 *     onClick={() => saveUser({ ... })}>
 *       Save User
 *   </button>
 *   {userSaveReq.isError && <div>Save error: {'' + userSaveReq.error}</div>}
 * </div>
 * ```
 */

function useAsyncResult(getPromise, useEffectOnChangeList) {
  if (useEffectOnChangeList === void 0) {
    useEffectOnChangeList = [];
  }

  var _useState = useState({
    count: 0
  }),
      promiseCounter = _useState[0];

  function setPromise(promise) {
    var thisCount = promiseCounter.count + 1;
    promiseCounter.count = thisCount;

    if (!promise) {
      setRes(AsyncResult.empty());
      return;
    }

    if (AsyncResult.isAsyncResult(promise)) {
      setRes(promise);
      return;
    }

    setRes(AsyncResult.pending());
    AsyncResult.wrap(promise).then(function (res) {
      if (promiseCounter.count != thisCount) return;
      setRes(res);
    });
  }

  var _addBind = function _addBind(p) {
    return _extends({}, p, {
      bind: setPromise,
      clear: function clear() {
        return setPromise(null);
      }
    });
  };

  var _useState2 = useState(_addBind(AsyncResult.empty())),
      res = _useState2[0],
      _setRes = _useState2[1];

  var setRes = function setRes(p) {
    return _setRes(_addBind(p));
  };

  useEffect(function () {
    setPromise(getPromise ? getPromise() : null);
  }, useEffectOnChangeList || []);
  return res;
}

export { AsyncResult, useAsyncResult };
//# sourceMappingURL=react-use-async-result.esm.js.map
